import logging
from .exceptions import DispatchException
from .instrumentation import instrument_client
from .models import RayureEventProcess
from .prio import PrioritySet
from .reconciliation import reconciliate_by_event
from django.utils import timezone

logger = logging.getLogger('rayures')
registry = PrioritySet()


def dispatch(event: 'Event'):
    name = event.type
    reconciliation = reconciliate_by_event(event, persist=True)
    obj = reconciliation.instance
    errors = []
    try:
        process = RayureEventProcess.objects.create(event=event)
        for func in registry[name]:
            with instrument_client() as subcalls:
                trace = process.log_trace(func, subcalls)
                try:
                    func(event, obj)
                except Exception as error:
                    proc_error = trace.log_error(error)
                    msg = (
                        f"failed to handle {proc_error.func} for {event.type}. "
                        f"please see {proc_error.id} for more details"
                    )
                    logger.error(msg, extra={
                        'event': event,
                        'error': error,
                        'event_process': process
                    })
                    errors.append((proc_error, error))
            if errors:
                raise DispatchException(
                    'Failed dispatching due to several errors',
                    process=process,
                    event=event,
                    proc_errors=errors)
    finally:
        process.status = 'failure' if errors else 'success'
        process.ended_at = timezone.now()
        process.save(update_fields=['status', 'ended_at', 'traces'])
    return process


def listen(name, *, func=None, position=100):
    """Decorator for dispatcher

    >>> @listen('my.event')
    >>> def func(event, obj):
    >>>     assert event.stripe_type == 'my.event'
    """
    def wrap(func):
        global registry
        # TODO: verify signature
        registry.add(name, func, position)
        return func
    if func:
        return wrap(func)
    return wrap



# Generated by Django 2.0.7 on 2018-07-10 11:05

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('rayures', '0002_orderreturn'),
    ]

    operations = [
        migrations.CreateModel(
            name='RayureMeta',
            fields=[
                ('id', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('deleted_at', models.DateTimeField(null=True)),
                ('idempotency_key', models.TextField(null=True)),
                ('request_id', models.TextField(null=True)),
                ('source', models.TextField(null=True)),
                ('content_type', models.ForeignKey(limit_choices_to=models.Q(app_label='rayures'), on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('event', models.ForeignKey(null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='rayures.Event')),
            ],
        ),
        migrations.RemoveField(
            model_name='rayuresmeta',
            name='content_type',
        ),
        migrations.RemoveField(
            model_name='rayuresmeta',
            name='event',
        ),
        migrations.DeleteModel(
            name='RayuresMeta',
        ),
    ]



from django.apps import AppConfig
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


class RayuresConfig(AppConfig):
    name = "rayures"

    def ready(self):
        from rayures import __version__
        import stripe
        for key in ("STRIPE_API_KEY", "STRIPE_ENDPOINT_SECRET", "STRIPE_PUBLISHABLE_KEY"):
            if not hasattr(settings, key):
                raise ImproperlyConfigured(f"{key} is mandatory")

        # configure stripe agent
        self.api_key = stripe.api_key = settings.STRIPE_API_KEY
        self.api_version = stripe.api_version = "2018-05-21"  # TODO: should it be configurable?
        stripe.set_app_info("django-rayures", version=__version__, url="https://lab.errorist.xyz/django/rayures")

        # client configurations
        self.endpoint_secret = getattr(settings, "STRIPE_ENDPOINT_SECRET", None)
        self.publishable_key = getattr(settings, "STRIPE_PUBLISHABLE_KEY", None)

        # Activate signals
        from . import tasks  # noqa

        # load webhook events from any project
        from django.utils.module_loading import autodiscover_modules
        autodiscover_modules('stripe_webhooks')



import factory
from . import models
from .objs import Price
from .utils import price_to_stripe
from datetime import datetime, timedelta


def timestamp(obj):
    if obj is None:
        return
    return int(obj.timestamp())


class CustomerFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Customer
        exclude = 'default_source', 'sources',

    id = factory.Faker('md5', raw_output=False)
    delinquent = False
    default_source = None
    sources = []

    data = factory.Dict({
        "object": "customer",
        "id": factory.SelfAttribute('..id'),
        "account_balance": 0,
        "currency": "usd",
        "delinquent": factory.SelfAttribute('..delinquent'),
        "default_source": factory.SelfAttribute('..default_source'),
        "sources": factory.Dict({
            "data": factory.SelfAttribute('...sources'),
        })
    })


class SubscriptionFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Subscription
        exclude = ('timestamp_current_period_start_at',
                   'timestamp_current_period_end_at',
                   'timestamp_start_at',
                   'now',
                   'timestamp_trial_start_at',
                   'timestamp_trial_end_at',
                   'timestamp_canceled_at',
                   'timestamp_ended_at')

    id = factory.Faker('md5', raw_output=False)
    customer = factory.SubFactory('rayures.factories.CustomerFactory')
    plan = factory.SubFactory('rayures.factories.PlanFactory')
    status = 'active'
    canceled_at = None
    ended_at = None

    now = factory.LazyFunction(datetime.utcnow)

    start_at = factory.LazyAttribute(lambda o: (o.now - timedelta(days=40)))
    current_period_start_at = factory.LazyAttribute(lambda o: (o.now - timedelta(days=1)))
    current_period_end_at = factory.LazyAttribute(lambda o: (o.now + timedelta(days=29)))
    trial_start_at = None
    trial_end_at = None

    timestamp_start_at = factory.LazyAttribute(lambda o: timestamp(o.start_at))
    timestamp_canceled_at = factory.LazyAttribute(lambda o: timestamp(o.canceled_at))
    timestamp_ended_at = factory.LazyAttribute(lambda o: timestamp(o.ended_at))
    timestamp_current_period_start_at = factory.LazyAttribute(lambda o: timestamp(o.current_period_start_at))
    timestamp_current_period_end_at = factory.LazyAttribute(lambda o: timestamp(o.current_period_end_at))

    timestamp_trial_start_at = factory.LazyAttribute(lambda o: timestamp(o.trial_start_at))
    timestamp_trial_end_at = factory.LazyAttribute(lambda o: timestamp(o.trial_end_at))
    cancel_at_period_end = False

    data = factory.Dict({
        "id": factory.SelfAttribute('..id'),
        "object": "subscription",
        "customer": factory.SelfAttribute('..customer.id'),
        "status": factory.SelfAttribute('..status'),
        "plan": factory.Dict({
            "id": factory.SelfAttribute('...plan.id'),
            "object": "plan",
        }),
        "items": factory.Dict({
            "object": "list",
            "data": factory.List([
                factory.Dict({
                    "id": factory.Faker('md5', raw_output=False),
                    "object": "subscription_item",
                    "plan": factory.Dict({
                        "id": factory.SelfAttribute('......plan.id'),
                        "object": "plan",
                    })
                })
            ])
        }),
        "current_period_start": factory.SelfAttribute('..timestamp_current_period_start_at'),
        "current_period_end": factory.SelfAttribute('..timestamp_current_period_end_at'),
        "cancel_at_period_end": factory.SelfAttribute('..cancel_at_period_end'),
        "start": factory.SelfAttribute('..timestamp_start_at'),
        "trial_start": factory.SelfAttribute('..timestamp_trial_start_at'),
        "trial_end": factory.SelfAttribute('..timestamp_trial_end_at'),
        "canceled_at": factory.SelfAttribute('..timestamp_canceled_at'),
        "ended_at": factory.SelfAttribute('..timestamp_ended_at'),
    })

    class Params:
        active = factory.Trait(
            status='active'
        )
        past_due = factory.Trait(
            status='past_due'
        )
        unpaid = factory.Trait(
            status='unpaid'
        )
        trialing = factory.Trait(
            status='trialing',
            start_at=factory.LazyAttribute(lambda o: (o.now - timedelta(days=40))),
            current_period_start_at=factory.LazyAttribute(lambda o: (o.now - timedelta(days=1))),
            current_period_end_at=factory.LazyAttribute(lambda o: (o.now - timedelta(days=6))),
            trial_start_at=factory.LazyAttribute(lambda o: o.current_period_start_at),
            trial_end_at=factory.LazyAttribute(lambda o: o.current_period_end_at),
        )
        canceling = factory.Trait(
            status='active',
            cancel_at_period_end=True
        )
        canceled = factory.Trait(
            status='canceled',
            start_at=factory.LazyAttribute(lambda o: (o.now - timedelta(days=40))),
            current_period_start_at=factory.LazyAttribute(lambda o: (o.now - timedelta(days=31))),
            current_period_end_at=factory.LazyAttribute(lambda o: (o.now - timedelta(days=1))),
            canceled_at=factory.LazyAttribute(lambda o: o.current_period_end_at),
            ended_at=factory.LazyAttribute(lambda o: o.current_period_end_at)
        )


class PlanFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Plan
        exclude = ('amount_value',
                   'amount_currency')

    id = factory.Faker('md5', raw_output=False)
    product = factory.SubFactory("rayures.factories.ProductFactory", type="service")

    amount = Price(100, 'usd')
    amount_currency = factory.LazyAttribute(lambda o: o.amount.currency)
    amount_value = factory.LazyAttribute(lambda o: price_to_stripe(o.amount)[0])

    interval = 'month'
    interval_count = 1

    active = True

    data = factory.Dict({
        "id": factory.SelfAttribute('..id'),
        "object": "plan",
        "product": factory.SelfAttribute('..product.id'),
        "amount": factory.SelfAttribute('..amount_value'),
        "currency": factory.SelfAttribute('..amount_currency'),
        "interval": factory.SelfAttribute('..interval'),
        "interval_count": factory.SelfAttribute('..interval_count'),
        "active": factory.SelfAttribute('..active'),
        "billing_scheme": 'per_unit'
    })


class ProductFactory(factory.Factory):
    class Meta:
        model = models.Product

    id = factory.Faker('md5', raw_output=False)
    type = "service"
    name = factory.Faker('sentence')
    caption = None
    description = None

    data = factory.Dict({
        "object": "product",
        "id": factory.SelfAttribute('..id'),
        "type": factory.SelfAttribute('..type'),
        "active": True,
        "attributes": [],
        "caption": factory.SelfAttribute('..caption'),
        "created": 1521455493,
        "deactivate_on": [],
        "description": factory.SelfAttribute('..description'),
        "images": [],
        "livemode": True,
        "metadata": {},
        "name": factory.SelfAttribute('..name'),
        "package_dimensions": None,
        "shippable": False,
        "updated": 1530793625,
        "url": None
    })

    class Params:
        good = factory.Trait(
            type="good",
            name=factory.Faker('sentence'),
            caption=factory.Faker('sentence'),
            description=factory.Faker('sentence')
        )
        service = factory.Trait(
            type="service",
            name=factory.Faker('sentence'),
            caption=None,
            description=None
        )


class CardFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Card

    id = factory.Faker('md5', raw_output=False)
    fingerprint = factory.Faker('md5', raw_output=False)
    customer = factory.SubFactory("rayures.factories.CustomerFactory")
    data = factory.Dict({
        "id": factory.SelfAttribute('..id'),
        "name": None,
        "brand": "Visa",
        "last4": "3956",
        "object": "card",
        "country": "US",
        "funding": "credit",
        "customer": factory.SelfAttribute('..customer.id'),
        "exp_year": 2018,
        "metadata": {},
        "cvc_check": "pass",
        "exp_month": 10,
        "fingerprint": factory.SelfAttribute('..fingerprint'),
        "tokenization_method": None
    })


class ChargeFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Charge

    id = factory.Faker('md5', raw_output=False)
    customer = factory.SubFactory('rayures.factories.CustomerFactory')
    metadata = factory.LazyFunction(dict)
    data = factory.Dict({
        "id": factory.SelfAttribute('..id'),
        "object": 'charge',
        "currency": 'usd',
        "amount": 666,
        "customer": factory.SelfAttribute('..customer.id'),
        "metadata": factory.SelfAttribute('..metadata'),
    })


class InvoiceFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Invoice

    id = factory.Faker('md5', raw_output=False)
    customer = factory.SubFactory('rayures.factories.CustomerFactory')
    subscription = factory.SubFactory('rayures.factories.SubscriptionFactory',
                                      customer=factory.SelfAttribute('..customer'))
    metadata = factory.LazyFunction(dict)
    data = factory.Dict({
        "id": factory.SelfAttribute('..id'),
        "object": 'invoice',
        "customer": factory.SelfAttribute('..customer.id'),
        "subscription": factory.SelfAttribute('..subscription.id'),
        "metadata": factory.SelfAttribute('..metadata'),
        "paid": True,
        "closed": True
    })


class EventFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Event
        exclude = ('now',
                   'timestamp_created_at',
                   'embedded_object')

    id = factory.Faker('md5', raw_output=False)
    type = 'customer.created'

    now = factory.LazyFunction(datetime.utcnow)

    created_at = factory.LazyAttribute(lambda o: (o.now - timedelta(days=40)))
    timestamp_created_at = factory.LazyAttribute(lambda o: timestamp(o.created_at))
    # embedded_object = factory.SubFactory('rayures.factories.CustomerFactory')
    content_object = factory.SubFactory('rayures.factories.CustomerFactory')
    # object_id = factory.SubFactory('rayures.factories.CustomerFactory')

    data = factory.Dict({
        "id": factory.SelfAttribute('..id'),
        "object": 'event',
        'type': factory.SelfAttribute('..type'),
        'api_version': '2018-05-21',
        "created": factory.SelfAttribute('..timestamp_created_at'),
        "data": factory.Dict({
            "object": factory.SelfAttribute('...content_object.data'),
            # "object": factory.SelfAttribute('...embedded_object.data'),
        }),
    })


class SKUFactory(factory.Factory):
    class Meta:
        model = models.SKU
        exclude = ('price_currency', 'price_value')

    id = factory.Faker('md5', raw_output=False)
    product = factory.SubFactory("rayures.factories.ProductFactory", type="good")

    price = Price(100, 'usd')
    price_currency = factory.LazyAttribute(lambda o: o.price.currency)
    price_value = factory.LazyAttribute(lambda o: price_to_stripe(o.price)[0])

    attributes = factory.LazyFunction(dict)
    data = factory.Dict({
        "id": factory.SelfAttribute('..id'),
        "object": 'sku',
        "image": None,
        "price": factory.SelfAttribute('..price_value'),
        "active": True,
        "created": 1513784110,
        "product": factory.SelfAttribute('..product.id'),
        "updated": 1516982164,
        "currency": factory.SelfAttribute('..price_currency'),
        "livemode": True,
        "metadata": {},
        "inventory": {
            "type": "infinite",
            "value": None,
            "quantity": None
        },
        "attributes": factory.SelfAttribute('..attributes'),
        "package_dimensions": None
    })



import stripe
from . import models
from .utils import dt_from_stripe
from collections import namedtuple
from django.contrib.contenttypes.models import ContentType
from django.db import transaction
from django.db.models.functions import Now
from django.utils import timezone
from email.utils import parsedate_to_datetime


def squeeze_state(state, *,
                  created=None,
                  deleted=None,
                  now=None,
                  api_version=None,
                  request_id=None,
                  idempotency_key=None) -> dict:
    opts = {
        'data': state,
        'created': created,
        'deleted': deleted,
        'now': now,
        'api_version': api_version,
        'request_id': request_id,
        'idempotency_key': idempotency_key,
        'id': None,
        'object': None,
        'account': None,
    }
    last_response = None
    last_response_date = None

    if isinstance(state, stripe.stripe_object.StripeObject):
        # bingo! best alternative
        last_response = state.last_response
        opts.update({
            'account': state.stripe_account,
            'object': getattr(type(state), "class_name", None),
            'id': state.get('id', None)
        })
        if state.stripe_version:
            opts["api_version"] = state.stripe_version
    elif isinstance(state, stripe.stripe_response.StripeResponse):
        last_response = state

    if last_response:
        opts.update({
            'data': last_response.data,
            'request_id': opts['request_id'] or last_response.request_id,
            'idempotency_key': last_response.idempotency_key,
            'api_version': opts['api_version'] or last_response.headers.get('Stripe-Version', None),
        })
        dt = last_response.headers.get('Date', None)
        if dt:
            opts['now'] = last_response_date = parsedate_to_datetime(dt)

    opts['object'] = opts['data'].get('object', opts['object'])  # no object smells bad
    opts['id'] = opts['data'].get('id', opts['id'])  # no id can be invoice.upcoming

    opts['deleted'] = deleted = opts['data'].get('deleted', deleted) is True
    if not deleted and opts['created'] is None:
        # let's try to find if it's a creation by comparing created attrs with Date header
        if ('created' in opts['data']) and last_response_date:
            dt1 = dt_from_stripe(state['created'])
            opts['created'] = dt1 == last_response_date

    return opts


def squeeze_event(event) -> dict:
    opts = {"source": "webhook"}
    if isinstance(event, dict):
        opts.update({
            "event_id": event.get("id", None)
        })
    if isinstance(event, stripe.Event):
        opts.update({
            "now": event["created"],  # FIXME: make it datetime,
            "api_version": event["api_version"],
            "state": event["data"]["object"],
            "event_id": event.id
        })
        if event["request"] is not None:
            opts.update({
                "request_id": event["request"]["id"],
                "idempotency_key": event["request"]["idempotency_key"],
            })
        event_type = event["type"]
    elif isinstance(event, models.Event):
        opts.update({
            "now": event.created_at,
            "api_version": event.api_version,
            "request_id": event.request_id,
            "idempotency_key": event.idempotency_key,
            "state": event.data["data"]["object"],
            "event_id": event.id
        })
        event_type = event.type
    else:
        raise NotImplementedError("whoot?")

    opts["created"] = event_type.endswith('.created')
    opts["deleted"] = event_type.endswith('.deleted')
    return opts


Reconciliation = namedtuple('Reconciliation', 'instance, persisted')


def reconciliate_by_event(event, *, persist=None) -> Reconciliation:
    opts = squeeze_event(event)
    opts.setdefault("persist", persist)

    # we are good so far
    rec = reconciliate(**opts)

    # TODO: link non persisted obj, like Discount and invoice.upcoming, customer.discount.created
    obj = rec.instance
    cls = type(obj)
    if rec.persisted and cls in models.PersistedModel.__subclasses__():
        event.content_type = ContentType.objects.get_for_model(cls)
        event.object_id = obj.id
        event.save(update_fields=['content_type', 'object_id'])
    return rec


def handle_volative_model(data, object, **opts):
    cls = {'balance': models.Balance, 'discount': models.Discount}.get(object)
    if cls:
        instance = cls(data)
        return Reconciliation(instance, False)
    raise NotImplementedError(f'Tristesse {object}')


def handle_incoming_model(data, object, now, deleted, created, api_version, **opts):
    cls = models.PERSISTED_MODELS.get(object)
    instance = cls(data=data, api_version=api_version)
    if created:  # stripe does not give a delete time, try to stick to almost near
        instance.created_at = now or timezone.now()
    if deleted:  # stripe does not give a delete time, try to stick to almost near
        instance.deleted_at = now or timezone.now()
    instance.rebound_fields()
    # TODO: do we care about meta?
    return Reconciliation(instance, False)


def handle_persistable_model(data, object, id, persist, api_version, now, deleted, created, **opts):
    cls = models.PERSISTED_MODELS.get(object)

    defaults = {
        'api_version': api_version
    }
    # 99.999% of the time, deletion are a total mess
    if not data.get('deleted', None):
        defaults['data'] = data

    def load_instance(stripe_id, defaults, cls, qs, created, deleted, now):
        instance = qs.filter(id=stripe_id).first()
        newborn = not instance
        if instance:
            for k, v in defaults.items():
                setattr(instance, k, v)
        else:
            instance = cls(id=stripe_id, **defaults)
            instance.created_at = now
        if created:  # stripe does not give always a create time, try to stick to almost near
            instance.created_at = getattr(instance, 'created_at', None) or now or timezone.now()
        if deleted:  # stripe does not give a delete time, try to stick to almost near
            instance.deleted_at = getattr(instance, 'deleted_at', None) or now or timezone.now()
        else:
            instance.updated_at = now
        instance.rebound_fields()
        return instance, newborn

    if persist is True:
        with transaction.atomic():
            qs = cls.objects.select_for_update(of=('self',))
            instance, newborn = load_instance(id, defaults, cls, qs, created, deleted, now)
            instance.save()
        return Reconciliation(instance, True)
    else:
        qs = cls.objects
        instance, newborn = load_instance(id, defaults, cls, qs, created, deleted, now)
        # TODO: do we care about meta?
        return Reconciliation(instance, False)


def reconciliate_event(state, *,
                       persist=None,
                       api_version=None,
                       request_id=None,
                       idempotency_key=None) -> Reconciliation:
    opts = squeeze_state(state,
                         api_version=api_version,
                         request_id=request_id,
                         idempotency_key=idempotency_key)
    opts.setdefault('persist', persist)
    opts.setdefault('source', "webhook")

    return handle_persistable_model(**opts)


def reconciliate(state, *,
                 created=None,
                 deleted=None,
                 persist=None,
                 source="sdk",
                 now=None,
                 api_version=None,
                 request_id=None,
                 idempotency_key=None,
                 event_id=None) -> Reconciliation:
    opts = squeeze_state(state,
                         created=created,
                         deleted=deleted,
                         now=now,
                         api_version=api_version,
                         request_id=request_id,
                         idempotency_key=idempotency_key)
    opts.setdefault('persist', persist)
    opts.setdefault('source', source)
    opts.setdefault('event_id', event_id)

    if opts['object'] not in models.PERSISTED_MODELS:
        rec = handle_volative_model(**opts)
    elif opts['id'] is None:
        rec = handle_incoming_model(**opts)
    else:
        rec = handle_persistable_model(**opts)

    # FIXME: handle meta now!
    if rec.persisted and opts['object'] not in ('event',):
        handle_meta(rec.instance, **opts)
    return rec


def handle_meta(instance, *, created, deleted, event_id, request_id, idempotency_key, source, **opts):
    cls = type(instance)
    content_type = ContentType.objects.get_for_model(cls)
    defaults = {
        'event_id': event_id,
        'request_id': request_id,
        'idempotency_key': idempotency_key,
        'source': source
    }
    if created:
        defaults['created_at'] = Now()
        defaults['deleted_at'] = None
    if deleted:
        defaults['deleted_at'] = Now()
    else:
        defaults['updated_at'] = Now()
    meta, created = models.RayureMeta.objects.update_or_create(id=instance.id,
                                                               content_type=content_type,
                                                               defaults=defaults)
    return meta



import decimal
import hmac
from .objs import Price
from datetime import datetime
from django.utils import timezone
from functools import singledispatch
from hashlib import sha256
from time import time
from typing import Tuple


# currencies those amount=1 means 100 cents
# https://support.stripe.com/questions/which-zero-decimal-currencies-does-stripe-support
ZERO_DECIMAL_CURRENCIES = [
    "bif", "clp", "djf", "gnf", "jpy", "kmf", "krw",
    "mga", "pyg", "rwf", "vuv", "xaf", "xof", "xpf",
]

CURRENCY_SYMBOLS = {
    "aud": "\u0024",
    "cad": "\u0024",
    "chf": "\u0043\u0048\u0046",
    "cny": "\u00a5",
    "eur": "\u20ac",
    "gbp": "\u00a3",
    "jpy": "\u00a5",
    "myr": "\u0052\u004d",
    "sgd": "\u0024",
    "usd": "\u0024",
}


def price_to_dict(price):
    return {
        'amount': price.amount,
        'currency': price.currency
    }


def price_from_stripe(amount, currency) -> Price:
    if (amount, currency) == (0, None):
        return Price(0, 'usd')
    if amount is not None:
        if currency.lower() not in ZERO_DECIMAL_CURRENCIES:
            amount = amount / decimal.Decimal("100")
        return Price(amount, currency)


def price_to_stripe(price: Price) -> Tuple[int, str]:
    amount = price.amount
    currency = price.currency
    if currency.lower() not in ZERO_DECIMAL_CURRENCIES:
        amount = amount * 100
    return int(amount), currency


@singledispatch
def dt_from_stripe(ts):
    return ts


@dt_from_stripe.register(int)
@dt_from_stripe.register(float)
def dt_from_stripe_number(ts):
    tz = timezone.utc
    return datetime.fromtimestamp(int(ts), tz)


@singledispatch
def dt_to_stripe(ts):
    return ts


@dt_to_stripe.register(datetime)
def dt_to_stripe_datetime(ts):
    return int(ts.timestamp())


def charge_now(ts, ref=None):
    ref = ref or timezone.now()
    if dt_from_stripe(ts) > ref:
        return 'now'
    return dt_to_stripe(ts)


def sign_request(payload, secret, timestamp=None):
    timestamp = int(timestamp or time())
    signed_payload = f"{timestamp}.{payload}".encode('utf-8')
    expected_signature = hmac \
        .new(secret.encode('utf-8'), msg=signed_payload, digestmod=sha256) \
        .hexdigest()
    signature = f"t={timestamp},v1={expected_signature}"
    return signature, payload, timestamp



from django.apps import apps
from django.core.management.base import BaseCommand
from rayures.events import dispatch


class Command(BaseCommand):
    help = 'Sync stripe events'

    def handle(self, *args, **options):
        # TODO: option to select only the one that have only failed or never processed
        cls = apps.get_model('rayures', 'Event')
        qs = cls.objects.all()
        # qs = qs.filter(type__startswith='coupon.')
        for event in qs.order_by('created_at'):
            print(event, event.created_at, event.type)
            dispatch(event)



# Generated by Django 2.0.7 on 2018-07-06 09:40

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rayures', '0001_squashed_0005_auto_20180630_2251'),
    ]

    operations = [
        migrations.CreateModel(
            name='OrderReturn',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



#!/usr/bin/env python3

from setuptools import setup, find_packages
import versioneer

setup(
    name="rayures",
    version=versioneer.get_version(),
    author="Xavier Barbosa",
    author_email="clint.northwood@gmail.com",
    url="https://lab.errorist.xyz/django/rayures",
    description="Integrate stripe into django application",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=[
        "stripe>=1.82.1"
    ],
    extras_require={
        'factories':  ["factory_boy>=2.11.1"],
    },
    license='MIT',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Content Management System",
        "Topic :: Office/Business :: Financial",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    cmdclass=versioneer.get_cmdclass()
)



from operator import itemgetter
from collections import defaultdict

__all__ = ['PrioritySet']


def key_generator(key):
    result = []
    parts = key.split('.')
    for parts in iterakey(key.split('.')):
        result.append('.'.join(parts))
    return sorted(result)


def iterakey(parts):
    if not parts:
        raise StopIteration
    first, *remains = parts
    if remains:
        for x in iterakey(remains):
            if x[0] == '*':
                # fold *.* to *
                yield x
            else:
                yield ('*', *x)
            yield (first, *x)
    else:
            yield '*',
            yield first,


class PrioritySet:
    def __init__(self):
        self.funcs = defaultdict(dict)

    def add(self, key, func, position=None):
        self.funcs[key][func] = position
        return func

    def __getitem__(self, key):
        funcs = {}
        for k in key_generator(key):
            funcs.update(self.funcs[k])
        funcs.update(self.funcs[key])
        return [func for func, _ in sorted(funcs.items(), key=itemgetter(1))]
